@@ -19,6 +19,7 @@ import collections
 import copy
 import json
 import os
+import re
 import sys
 import warnings
 
@@ -68,6 +69,8 @@ VENDOR_FILES = [
 
 BOOL_KEYS = ('insecure', 'cache')
 
+FORMAT_EXCLUSIONS = frozenset(['password'])
+
 
 # NOTE(dtroyer): This turns out to be not the best idea so let's move
 #                overriding defaults to a kwarg to OpenStackConfig.__init__()
@@ -83,6 +86,8 @@ def set_default(key, value):
 
 
 def get_boolean(value):
+    if value is None:
+        return False
     if type(value) is bool:
         return value
     if value.lower() == 'true':
@@ -103,12 +108,14 @@ def _get_os_environ(envvar_prefix=None):
     for k in environkeys:
         newkey = k.split('_', 1)[-1].lower()
         ret[newkey] = os.environ[k]
-    # If the only environ key is region name, don't make a cloud, because
-    # it's being used as a cloud selector
-    if not environkeys or (
-            len(environkeys) == 1 and 'region_name' in ret):
-        return None
-    return ret
+    # If the only environ keys are selectors or behavior modification, don't
+    # return anything
+    selectors = set([
+        'OS_CLOUD', 'OS_REGION_NAME',
+        'OS_CLIENT_CONFIG_FILE', 'OS_CLIENT_SECURE_FILE', 'OS_CLOUD_NAME'])
+    if set(environkeys) - selectors:
+        return ret
+    return None
 
 
 def _merge_clouds(old_dict, new_dict):
@@ -145,7 +152,9 @@ def _fix_argv(argv):
     # over the place.
     processed = collections.defaultdict(list)
     for index in range(0, len(argv)):
-        if argv[index].startswith('--'):
+        # If the value starts with '--' and has '-' or '_' in it, then
+        # it's worth looking at it
+        if re.match('^--.*(_|-)+.*', argv[index]):
             split_args = argv[index].split('=')
             orig = split_args[0]
             new = orig.replace('_', '-')
@@ -171,18 +180,29 @@ class OpenStackConfig(object):
 
     def __init__(self, config_files=None, vendor_files=None,
                  override_defaults=None, force_ipv4=None,
-                 envvar_prefix=None, secure_files=None):
+                 envvar_prefix=None, secure_files=None,
+                 pw_func=None, session_constructor=None,
+                 app_name=None, app_version=None,
+                 load_yaml_config=True):
         self.log = _log.setup_logging(__name__)
+        self._session_constructor = session_constructor
+        self._app_name = app_name
+        self._app_version = app_version
 
-        self._config_files = config_files or CONFIG_FILES
-        self._secure_files = secure_files or SECURE_FILES
-        self._vendor_files = vendor_files or VENDOR_FILES
+        if load_yaml_config:
+            self._config_files = config_files or CONFIG_FILES
+            self._secure_files = secure_files or SECURE_FILES
+            self._vendor_files = vendor_files or VENDOR_FILES
+        else:
+            self._config_files = []
+            self._secure_files = []
+            self._vendor_files = []
 
-        config_file_override = os.environ.pop('OS_CLIENT_CONFIG_FILE', None)
+        config_file_override = os.environ.get('OS_CLIENT_CONFIG_FILE')
         if config_file_override:
             self._config_files.insert(0, config_file_override)
 
-        secure_file_override = os.environ.pop('OS_CLIENT_SECURE_FILE', None)
+        secure_file_override = os.environ.get('OS_CLIENT_SECURE_FILE')
         if secure_file_override:
             self._secure_files.insert(0, secure_file_override)
 
@@ -211,12 +231,12 @@ class OpenStackConfig(object):
         else:
             # Get the backwards compat value
             prefer_ipv6 = get_boolean(
-                os.environ.pop(
+                os.environ.get(
                     'OS_PREFER_IPV6', client_config.get(
                         'prefer_ipv6', client_config.get(
                             'prefer-ipv6', True))))
             force_ipv4 = get_boolean(
-                os.environ.pop(
+                os.environ.get(
                     'OS_FORCE_IPV4', client_config.get(
                         'force_ipv4', client_config.get(
                             'broken-ipv6', False))))
@@ -228,7 +248,7 @@ class OpenStackConfig(object):
                 self.force_ipv4 = True
 
         # Next, process environment variables and add them to the mix
-        self.envvar_key = os.environ.pop('OS_CLOUD_NAME', 'envvars')
+        self.envvar_key = os.environ.get('OS_CLOUD_NAME', 'envvars')
         if self.envvar_key in self.cloud_config['clouds']:
             raise exceptions.OpenStackConfigException(
                 '"{0}" defines a cloud named "{1}", but'
@@ -236,9 +256,8 @@ class OpenStackConfig(object):
                 ' either your environment based cloud, or one of your'
                 ' file-based clouds.'.format(self.config_filename,
                                              self.envvar_key))
-        # Pull out OS_CLOUD so that if it's the only thing set, do not
-        # make an envvars cloud
-        self.default_cloud = os.environ.pop('OS_CLOUD', None)
+
+        self.default_cloud = os.environ.get('OS_CLOUD')
 
         envvars = _get_os_environ(envvar_prefix=envvar_prefix)
         if envvars:
@@ -246,6 +265,17 @@ class OpenStackConfig(object):
             if not self.default_cloud:
                 self.default_cloud = self.envvar_key
 
+        if not self.default_cloud and self.cloud_config['clouds']:
+            if len(self.cloud_config['clouds'].keys()) == 1:
+                # If there is only one cloud just use it. This matches envvars
+                # behavior and allows for much less typing.
+                # TODO(mordred) allow someone to mark a cloud as "default" in
+                # clouds.yaml.
+                # The next/iter thing is for python3 compat where dict.keys
+                # returns an iterator but in python2 it's a list.
+                self.default_cloud = next(iter(
+                    self.cloud_config['clouds'].keys()))
+
         # Finally, fall through and make a cloud that starts with defaults
         # because we need somewhere to put arguments, and there are neither
         # config files or env vars
@@ -286,6 +316,10 @@ class OpenStackConfig(object):
         # Flag location to hold the peeked value of an argparse timeout value
         self._argv_timeout = False
 
+        # Save the password callback
+        # password = self._pw_callback(prompt="Password: ")
+        self._pw_callback = pw_func
+
     def get_extra_config(self, key, defaults=None):
         """Fetch an arbitrary extra chunk of config, laying in defaults.
 
@@ -381,7 +415,10 @@ class OpenStackConfig(object):
         if 'regions' in config:
             return self._expand_regions(config['regions'])
         elif 'region_name' in config:
-            regions = config['region_name'].split(',')
+            if isinstance(config['region_name'], list):
+                regions = config['region_name']
+            else:
+                regions = config['region_name'].split(',')
             if len(regions) > 1:
                 warnings.warn(
                     "Comma separated lists in region_name are deprecated."
@@ -432,7 +469,7 @@ class OpenStackConfig(object):
         # Only validate cloud name if one was given
         if name and name not in self.cloud_config['clouds']:
             raise exceptions.OpenStackConfigException(
-                "Named cloud {name} requested that was not found.".format(
+                "Cloud {name} was not found.".format(
                     name=name))
 
         our_cloud = self.cloud_config['clouds'].get(name, dict())
@@ -466,6 +503,17 @@ class OpenStackConfig(object):
             else:
                 profile_data = vendors.get_profile(profile_name)
                 if profile_data:
+                    status = profile_data.pop('status', 'active')
+                    message = profile_data.pop('message', '')
+                    if status == 'deprecated':
+                        warnings.warn(
+                            "{profile_name} is deprecated: {message}".format(
+                                profile_name=profile_name, message=message))
+                    elif status == 'shutdown':
+                        raise exceptions.OpenStackConfigException(
+                            "{profile_name} references a cloud that no longer"
+                            " exists: {message}".format(
+                                profile_name=profile_name, message=message))
                     _auth_update(cloud, profile_data)
                 else:
                     # Can't find the requested vendor config, go about business
@@ -473,18 +521,77 @@ class OpenStackConfig(object):
                                   " the cloud '{1}'".format(profile_name,
                                                             name))
 
-    def _fix_backwards_madness(self, cloud):
-        cloud = self._fix_backwards_auth_plugin(cloud)
-        cloud = self._fix_backwards_project(cloud)
-        cloud = self._fix_backwards_interface(cloud)
-        cloud = self._handle_domain_id(cloud)
-        return cloud
-
     def _project_scoped(self, cloud):
         return ('project_id' in cloud or 'project_name' in cloud
                 or 'project_id' in cloud['auth']
                 or 'project_name' in cloud['auth'])
 
+    def _validate_networks(self, networks, key):
+        value = None
+        for net in networks:
+            if value and net[key]:
+                raise exceptions.OpenStackConfigException(
+                    "Duplicate network entries for {key}: {net1} and {net2}."
+                    " Only one network can be flagged with {key}".format(
+                        key=key,
+                        net1=value['name'],
+                        net2=net['name']))
+            if not value and net[key]:
+                value = net
+
+    def _fix_backwards_networks(self, cloud):
+        # Leave the external_network and internal_network keys in the
+        # dict because consuming code might be expecting them.
+        networks = []
+        # Normalize existing network entries
+        for net in cloud.get('networks', []):
+            name = net.get('name')
+            if not name:
+                raise exceptions.OpenStackConfigException(
+                    'Entry in network list is missing required field "name".')
+            network = dict(
+                name=name,
+                routes_externally=get_boolean(net.get('routes_externally')),
+                nat_source=get_boolean(net.get('nat_source')),
+                nat_destination=get_boolean(net.get('nat_destination')),
+                default_interface=get_boolean(net.get('default_interface')),
+            )
+            # routes_ipv4_externally defaults to the value of routes_externally
+            network['routes_ipv4_externally'] = get_boolean(
+                net.get(
+                    'routes_ipv4_externally', network['routes_externally']))
+            # routes_ipv6_externally defaults to the value of routes_externally
+            network['routes_ipv6_externally'] = get_boolean(
+                net.get(
+                    'routes_ipv6_externally', network['routes_externally']))
+            networks.append(network)
+
+        for key in ('external_network', 'internal_network'):
+            external = key.startswith('external')
+            if key in cloud and 'networks' in cloud:
+                raise exceptions.OpenStackConfigException(
+                    "Both {key} and networks were specified in the config."
+                    " Please remove {key} from the config and use the network"
+                    " list to configure network behavior.".format(key=key))
+            if key in cloud:
+                warnings.warn(
+                    "{key} is deprecated. Please replace with an entry in"
+                    " a dict inside of the networks list with name: {name}"
+                    " and routes_externally: {external}".format(
+                        key=key, name=cloud[key], external=external))
+                networks.append(dict(
+                    name=cloud[key],
+                    routes_externally=external,
+                    nat_destination=not external,
+                    default_interface=external))
+
+        # Validate that we don't have duplicates
+        self._validate_networks(networks, 'nat_destination')
+        self._validate_networks(networks, 'default_interface')
+
+        cloud['networks'] = networks
+        return cloud
+
     def _handle_domain_id(self, cloud):
         # Allow people to just specify domain once if it's the same
         mappings = {
@@ -499,6 +606,7 @@ class OpenStackConfig(object):
             for key in possible_values:
                 if target_key in cloud['auth'] and key not in cloud['auth']:
                     cloud['auth'][key] = cloud['auth'][target_key]
+            cloud.pop(target_key, None)
             cloud['auth'].pop(target_key, None)
         return cloud
 
@@ -560,13 +668,14 @@ class OpenStackConfig(object):
         # completely broken
         return cloud
 
-    def register_argparse_arguments(self, parser, argv, service_keys=[]):
+    def register_argparse_arguments(self, parser, argv, service_keys=None):
         """Register all of the common argparse options needed.
 
         Given an argparse parser, register the keystoneauth Session arguments,
         the keystoneauth Auth Plugin Options and os-cloud. Also, peek in the
         argv to see if all of the auth plugin options should be registered
         or merely the ones already configured.
+
         :param argparse.ArgumentParser: parser to attach argparse options to
         :param list argv: the arguments provided to the application
         :param string service_keys: Service or list of services this argparse
@@ -579,6 +688,9 @@ class OpenStackConfig(object):
                                                     is requested
         """
 
+        if service_keys is None:
+            service_keys = []
+
         # Fix argv in place - mapping any keys with embedded _ in them to -
         _fix_argv(argv)
 
@@ -703,7 +815,7 @@ class OpenStackConfig(object):
                         cloud, region_name=region['name']))
         return clouds
 
-    def _fix_args(self, args, argparse=None):
+    def _fix_args(self, args=None, argparse=None):
         """Massage the passed-in options
 
         Replace - with _ and strip os_ prefixes.
@@ -711,6 +823,8 @@ class OpenStackConfig(object):
         Convert an argparse Namespace object to a dict, removing values
         that are either None or ''.
         """
+        if not args:
+            args = {}
 
         if argparse:
             # Convert the passed-in Namespace
@@ -752,7 +866,7 @@ class OpenStackConfig(object):
     def auth_config_hook(self, config):
         """Allow examination of config values before loading auth plugin
 
-        OpenStackClient will override this to perform additional chacks
+        OpenStackClient will override this to perform additional checks
         on auth_type.
         """
         return config
@@ -769,56 +883,13 @@ class OpenStackConfig(object):
             # Set to notused rather than None because validate_auth will
             # strip the value if it's actually python None
             config['auth']['token'] = 'notused'
+        elif config['auth_type'] == 'token_endpoint':
+            # Humans have been trained to use a thing called token_endpoint
+            # That it does not exist in keystoneauth is irrelvant- it not
+            # doing what they want causes them sorrow.
+            config['auth_type'] = 'admin_token'
         return loading.get_plugin_loader(config['auth_type'])
 
-    def _validate_auth_ksc(self, config):
-        try:
-            import keystoneclient.auth as ksc_auth
-        except ImportError:
-            return config
-
-        # May throw a keystoneclient.exceptions.NoMatchingPlugin
-        plugin_options = ksc_auth.get_plugin_class(
-            config['auth_type']).get_options()
-
-        for p_opt in plugin_options:
-            # if it's in config.auth, win, kill it from config dict
-            # if it's in config and not in config.auth, move it
-            # deprecated loses to current
-            # provided beats default, deprecated or not
-            winning_value = self._find_winning_auth_value(
-                p_opt, config['auth'])
-            if not winning_value:
-                winning_value = self._find_winning_auth_value(p_opt, config)
-
-            # if the plugin tells us that this value is required
-            # then error if it's doesn't exist now
-            if not winning_value and p_opt.required:
-                raise exceptions.OpenStackConfigException(
-                    'Unable to find auth information for cloud'
-                    ' {cloud} in config files {files}'
-                    ' or environment variables. Missing value {auth_key}'
-                    ' required for auth plugin {plugin}'.format(
-                        cloud=cloud, files=','.join(self._config_files),
-                        auth_key=p_opt.name, plugin=config.get('auth_type')))
-
-            # Clean up after ourselves
-            for opt in [p_opt.name] + [o.name for o in p_opt.deprecated_opts]:
-                opt = opt.replace('-', '_')
-                config.pop(opt, None)
-                config['auth'].pop(opt, None)
-
-            if winning_value:
-                # Prefer the plugin configuration dest value if the value's key
-                # is marked as depreciated.
-                if p_opt.dest is None:
-                    config['auth'][p_opt.name.replace('-', '_')] = (
-                        winning_value)
-                else:
-                    config['auth'][p_opt.dest] = winning_value
-
-        return config
-
     def _validate_auth(self, config, loader):
         # May throw a keystoneauth1.exceptions.NoMatchingPlugin
 
@@ -830,25 +901,127 @@ class OpenStackConfig(object):
             # deprecated loses to current
             # provided beats default, deprecated or not
             winning_value = self._find_winning_auth_value(
-                p_opt, config['auth'])
+                p_opt,
+                config['auth'],
+            )
             if not winning_value:
-                winning_value = self._find_winning_auth_value(p_opt, config)
+                winning_value = self._find_winning_auth_value(
+                    p_opt,
+                    config,
+                )
 
-            # Clean up after ourselves
-            for opt in [p_opt.name] + [o.name for o in p_opt.deprecated]:
-                opt = opt.replace('-', '_')
-                config.pop(opt, None)
-                config['auth'].pop(opt, None)
+            config = self._clean_up_after_ourselves(
+                config,
+                p_opt,
+                winning_value,
+            )
 
             if winning_value:
                 # Prefer the plugin configuration dest value if the value's key
-                # is marked as depreciated.
+                # is marked as deprecated.
                 if p_opt.dest is None:
-                    config['auth'][p_opt.name.replace('-', '_')] = (
-                        winning_value)
+                    good_name = p_opt.name.replace('-', '_')
+                    config['auth'][good_name] = winning_value
                 else:
                     config['auth'][p_opt.dest] = winning_value
 
+            # See if this needs a prompting
+            config = self.option_prompt(config, p_opt)
+
+        return config
+
+    def _validate_auth_correctly(self, config, loader):
+        # May throw a keystoneauth1.exceptions.NoMatchingPlugin
+
+        plugin_options = loader.get_options()
+
+        for p_opt in plugin_options:
+            # if it's in config, win, move it and kill it from config dict
+            # if it's in config.auth but not in config it's good
+            # deprecated loses to current
+            # provided beats default, deprecated or not
+            winning_value = self._find_winning_auth_value(
+                p_opt,
+                config,
+            )
+            if not winning_value:
+                winning_value = self._find_winning_auth_value(
+                    p_opt,
+                    config['auth'],
+                )
+
+            config = self._clean_up_after_ourselves(
+                config,
+                p_opt,
+                winning_value,
+            )
+
+            # See if this needs a prompting
+            config = self.option_prompt(config, p_opt)
+
+        return config
+
+    def option_prompt(self, config, p_opt):
+        """Prompt user for option that requires a value"""
+        if (
+                getattr(p_opt, 'prompt', None) is not None and
+                p_opt.dest not in config['auth'] and
+                self._pw_callback is not None
+        ):
+            config['auth'][p_opt.dest] = self._pw_callback(p_opt.prompt)
+        return config
+
+    def _clean_up_after_ourselves(self, config, p_opt, winning_value):
+
+        # Clean up after ourselves
+        for opt in [p_opt.name] + [o.name for o in p_opt.deprecated]:
+            opt = opt.replace('-', '_')
+            config.pop(opt, None)
+            config['auth'].pop(opt, None)
+
+        if winning_value:
+            # Prefer the plugin configuration dest value if the value's key
+            # is marked as depreciated.
+            if p_opt.dest is None:
+                config['auth'][p_opt.name.replace('-', '_')] = (
+                    winning_value)
+            else:
+                config['auth'][p_opt.dest] = winning_value
+        return config
+
+    def magic_fixes(self, config):
+        """Perform the set of magic argument fixups"""
+
+        # Infer token plugin if a token was given
+        if (('auth' in config and 'token' in config['auth']) or
+                ('auth_token' in config and config['auth_token']) or
+                ('token' in config and config['token'])):
+            config.setdefault('token', config.pop('auth_token', None))
+
+        # These backwards compat values are only set via argparse. If it's
+        # there, it's because it was passed in explicitly, and should win
+        config = self._fix_backwards_api_timeout(config)
+        if 'endpoint_type' in config:
+            config['interface'] = config.pop('endpoint_type')
+
+        config = self._fix_backwards_auth_plugin(config)
+        config = self._fix_backwards_project(config)
+        config = self._fix_backwards_interface(config)
+        config = self._fix_backwards_networks(config)
+        config = self._handle_domain_id(config)
+
+        for key in BOOL_KEYS:
+            if key in config:
+                if type(config[key]) is not bool:
+                    config[key] = get_boolean(config[key])
+
+        # TODO(mordred): Special casing auth_url here. We should
+        #                come back to this betterer later so that it's
+        #                more generalized
+        if 'auth' in config and 'auth_url' in config['auth']:
+            config['auth']['auth_url'] = config['auth']['auth_url'].format(
+                **config)
+
         return config
 
     def get_one_cloud(self, cloud=None, validate=True,
@@ -901,31 +1074,8 @@ class OpenStackConfig(object):
                     else:
                         config[key] = val
 
-        # Infer token plugin if a token was given
-        if (('auth' in config and 'token' in config['auth']) or
-                ('auth_token' in config and config['auth_token']) or
-                ('token' in config and config['token'])):
-            config.setdefault('token', config.pop('auth_token', None))
-
-        # These backwards compat values are only set via argparse. If it's
-        # there, it's because it was passed in explicitly, and should win
-        config = self._fix_backwards_api_timeout(config)
-        if 'endpoint_type' in config:
-            config['interface'] = config.pop('endpoint_type')
-
-        config = self._fix_backwards_madness(config)
-
-        for key in BOOL_KEYS:
-            if key in config:
-                if type(config[key]) is not bool:
-                    config[key] = get_boolean(config[key])
-
-        # TODO(mordred): Special casing auth_url here. We should
-        #                come back to this betterer later so that it's
-        #                more generalized
-        if 'auth' in config and 'auth_url' in config['auth']:
-            config['auth']['auth_url'] = config['auth']['auth_url'].format(
-                **config)
+        config = self.magic_fixes(config)
+        config = self._normalize_keys(config)
 
         # NOTE(dtroyer): OSC needs a hook into the auth args before the
         #                plugin is loaded in order to maintain backward-
@@ -933,27 +1083,15 @@ class OpenStackConfig(object):
         config = self.auth_config_hook(config)
 
         if validate:
-            try:
-                loader = self._get_auth_loader(config)
-                config = self._validate_auth(config, loader)
-                auth_plugin = loader.load_from_options(**config['auth'])
-            except Exception as e:
-                # We WANT the ksa exception normally
-                # but OSC can't handle it right now, so we try deferring
-                # to ksc. If that ALSO fails, it means there is likely
-                # a deeper issue, so we assume the ksa error was correct
-                self.log.debug("Deferring keystone exception: {e}".format(e=e))
-                auth_plugin = None
-                try:
-                    config = self._validate_auth_ksc(config)
-                except Exception:
-                    raise e
+            loader = self._get_auth_loader(config)
+            config = self._validate_auth(config, loader)
+            auth_plugin = loader.load_from_options(**config['auth'])
         else:
             auth_plugin = None
 
         # If any of the defaults reference other values, we need to expand
         for (key, value) in config.items():
-            if hasattr(value, 'format'):
+            if hasattr(value, 'format') and key not in FORMAT_EXCLUSIONS:
                 config[key] = value.format(**config)
 
         force_ipv4 = config.pop('force_ipv4', self.force_ipv4)
@@ -966,12 +1104,108 @@ class OpenStackConfig(object):
         else:
             cloud_name = str(cloud)
         return cloud_config.CloudConfig(
-            name=cloud_name, region=config['region_name'],
+            name=cloud_name,
+            region=config['region_name'],
+            config=config,
+            force_ipv4=force_ipv4,
+            auth_plugin=auth_plugin,
+            openstack_config=self,
+            session_constructor=self._session_constructor,
+            app_name=self._app_name,
+            app_version=self._app_version,
+        )
+
+    def get_one_cloud_osc(
+        self,
+        cloud=None,
+        validate=True,
+        argparse=None,
+        **kwargs
+    ):
+        """Retrieve a single cloud configuration and merge additional options
+
+        :param string cloud:
+            The name of the configuration to load from clouds.yaml
+        :param boolean validate:
+            Validate the config. Setting this to False causes no auth plugin
+            to be created. It's really only useful for testing.
+        :param Namespace argparse:
+            An argparse Namespace object; allows direct passing in of
+            argparse options to be added to the cloud config.  Values
+            of None and '' will be removed.
+        :param region_name: Name of the region of the cloud.
+        :param kwargs: Additional configuration options
+
+        :raises: keystoneauth1.exceptions.MissingRequiredOptions
+            on missing required auth parameters
+        """
+
+        args = self._fix_args(kwargs, argparse=argparse)
+
+        if cloud is None:
+            if 'cloud' in args:
+                cloud = args['cloud']
+            else:
+                cloud = self.default_cloud
+
+        config = self._get_base_cloud_config(cloud)
+
+        # Get region specific settings
+        if 'region_name' not in args:
+            args['region_name'] = ''
+        region = self._get_region(cloud=cloud, region_name=args['region_name'])
+        args['region_name'] = region['name']
+        region_args = copy.deepcopy(region['values'])
+
+        # Regions is a list that we can use to create a list of cloud/region
+        # objects. It does not belong in the single-cloud dict
+        config.pop('regions', None)
+
+        # Can't just do update, because None values take over
+        for arg_list in region_args, args:
+            for (key, val) in iter(arg_list.items()):
+                if val is not None:
+                    if key == 'auth' and config[key] is not None:
+                        config[key] = _auth_update(config[key], val)
+                    else:
+                        config[key] = val
+
+        config = self.magic_fixes(config)
+
+        # NOTE(dtroyer): OSC needs a hook into the auth args before the
+        #                plugin is loaded in order to maintain backward-
+        #                compatible behaviour
+        config = self.auth_config_hook(config)
+
+        if validate:
+            loader = self._get_auth_loader(config)
+            config = self._validate_auth_correctly(config, loader)
+            auth_plugin = loader.load_from_options(**config['auth'])
+        else:
+            auth_plugin = None
+
+        # If any of the defaults reference other values, we need to expand
+        for (key, value) in config.items():
+            if hasattr(value, 'format') and key not in FORMAT_EXCLUSIONS:
+                config[key] = value.format(**config)
+
+        force_ipv4 = config.pop('force_ipv4', self.force_ipv4)
+        prefer_ipv6 = config.pop('prefer_ipv6', True)
+        if not prefer_ipv6:
+            force_ipv4 = True
+
+        if cloud is None:
+            cloud_name = ''
+        else:
+            cloud_name = str(cloud)
+        return cloud_config.CloudConfig(
+            name=cloud_name,
+            region=config['region_name'],
             config=self._normalize_keys(config),
             force_ipv4=force_ipv4,
             auth_plugin=auth_plugin,
-            openstack_config=self
-            )
+            openstack_config=self,
+        )
 
     @staticmethod
     def set_one_cloud(config_file, cloud, set_config=None):
